=head1 NAME

Slapd::Backend::Proxy - LDAP slapd proxy backend class

=head1 DESCRIPTION

Proxy synchronization between LDAP and local database.

=cut

package Slapd::Backend::Proxy;

use utf8;
use strict;
use warnings;

use IO::String;
use List::MoreUtils qw( any );
use Net::LDAP::Constant qw/
	LDAP_SUCCESS
	LDAP_COMPARE_FALSE
	LDAP_COMPARE_TRUE
	LDAP_INVALID_SYNTAX
	LDAP_NO_SUCH_OBJECT
	LDAP_UNWILLING_TO_PERFORM
	LDAP_NOT_ALLOWED_ON_RDN
/;
use Net::LDAP::Entry;
use Net::LDAP::LDIF;
use Sys::Syslog qw( LOG_INFO LOG_WARNING LOG_ERR LOG_DEBUG );

use Slapd::Backend::Constant;

use base 'Class::Accessor::Lvalue::Fast';

__PACKAGE__->mk_accessors(qw/ store verbose /);

$SIG{__WARN__} = \&_sig_warn;
$SIG{__DIE__} = \&_sig_die;

=head1 PUBLIC METHODS

=over

=item B<new>()

Purpose:
  * Constructor

Arguments:
  * None

Returns:
  * Slapd::Backend::Proxy object

=cut

sub new
{
	my $class = shift;

	my $self = bless {}, $class;

	$self->verbose = 1 if $ENV{SLAPD_PROXY_VERBOSE};

	$self->_info( "LDAP perl backend (%s) started", $class ) if $self->verbose;

	return $self;
}

=item B<init>()

Purpose:
  * Initialization

Arguments:
  * None

Returns:
  * Nonzero if initialization failed

=cut

sub init
{
	my $self = shift;
	$self->_debug( q/Init/ ) if $self->verbose;

	my $store_mod = sprintf "%s::Store::%s", __PACKAGE__, $self->{store_mod} ||= 'File';
	die $@ unless eval "require $store_mod";
	$self->store = $store_mod->new( @{ $self->{store_opt} ||= [] } );

	return $self->_init();
}

=item B<search>( $base, $scope, $deref, $size_lim, $time_lim, $filter_str, $attr_only, @attrs )

Purpose:
  * Search

Arguments:
  * Scalar - base DN
  * Scalar - scope
  * Scalar - alias dereferencing policy
  * Scalar - size limit
  * Scalar - time limit
  * Scalar - filter string
  * Scalar - attributes only flag
  * Array  - attributes (optional)

Returns:
  * Scalar - result code
  * Array  - LDIF entries

=cut

sub search
{
	my $self = shift;
	my( $base, $scope, $deref, $size_lim, $time_lim, $filter_str, $attr_only, @attrs ) = @_;

	$self->_debug(
		q/Search: base="%s", scope=%s, deref=%s, size_lim=%s, time_lim=%s, filter="%s", attr_only=%s, attrs=|%s|/,
		$base, $scope, $deref, $size_lim, $time_lim, $filter_str, $attr_only, join( '|', @attrs )
	) if $self->verbose;

	my %filter = %{$self->parse_filter( $filter_str )};
	return LDAP_INVALID_SYNTAX unless %filter;

	my( %attrs, %where );
	foreach my $attr ( keys %filter )
	{
		if( any { $attr eq $_ } $self->key_attribute )
		{
			$attrs{$attr} = $filter{$attr};
			$where{$attr} = $filter{$attr};
		}
		elsif( any { $attr eq $_ } $self->stored_attributes )
		{
			$where{$attr} = $filter{$attr};
		}
		else
		{
			$attrs{$attr} = $filter{$attr};
		}
	}

	# Search in entries
	my( @stored_entries, @match_entries );
	foreach my $entry ( $self->store->entries( $base, $scope, $filter{entryUUID} ) )
	{
		my( $ftr, $entries );
		if( $self->_is_entry_stored( $entry ) )
		{
			$ftr = \%attrs;
			$entries = \@stored_entries;
		}
		else
		{
			$ftr = \%filter;
			$entries = \@match_entries;
		}
		push @$entries, $entry if $self->check_filter( $entry, $ftr );
	}

	# Search in DB
	if( @stored_entries )
	{
		my( $code, @match ) = $self->_search( filter => \%where, entries => \@stored_entries );
		return $code if $code;
		push @match_entries, @match;
	}

	# If data was modified by dirty hands, we ought to delete entryCSN to force entry update
	foreach my $entry ( @match_entries )
	{
		my $ok = 1;
		foreach my $attr ( $self->stored_attributes )
		{
			my $ientry = $self->store->entry( $entry->dn() );
			$ok = undef unless [ sort $entry->get_value($attr) ] ~~ [ sort $ientry->get_value($attr) ];
		}
		$entry->delete( 'entryCSN' => undef ) unless $ok;
	}

	my @ldif_entries;
	foreach my $entry ( @match_entries )
	{
		my $str = new IO::String;
		my $ldif = new Net::LDAP::LDIF( $str, 'w', raw => qr/;binary/ );
		# write_entry() converts string to UTF-8 octets in-place, so we ought to clone entry.
		$ldif->write_entry( $entry->clone() );
		my $ldif_str = ${$str->string_ref()};
		$ldif_str =~ s/^\n//;
		push @ldif_entries, $ldif_str;
		last if $size_lim != -1 && @ldif_entries >= $size_lim;
	}

	return ( LDAP_SUCCESS, @ldif_entries );
}

=item B<compare>( $dn, $ava_str )

Purpose:
  * Compare

Arguments:
  * Scalar - DN
  * Scalar - attribute assertion string

Returns:
  * Scalar - result code

=cut

sub compare
{
	my $self = shift;
	my( $dn, $ava_str ) = @_;
	$self->_debug( q/Compare: dn="%s", ava_str="%s"/, $dn, $ava_str ) if $self->verbose;

	my( $attr, $value ) = split /=/, $ava_str, 2;
	$value = $self->_unescape( $value );

	my $entry = $self->store->entry( $dn );
	return LDAP_NO_SUCH_OBJECT unless defined $entry;

	if( $self->_is_entry_stored( $entry ) && any { $attr eq $_ } $self->stored_attributes )
	{
		my( $code, @entries ) = $self->_search( entries => [ $entry ] );
		return $code if $code != LDAP_SUCCESS;
		return LDAP_NO_SUCH_OBJECT unless @entries;
		$entry = shift @entries;
	}

	return lc $value ~~ [ map lc $_, $entry->get_value($attr) ]
		? LDAP_COMPARE_TRUE : LDAP_COMPARE_FALSE;
}

=item B<modify>( $dn, @list )

Purpose:
  * Modify

Arguments:
  * Scalar - DN
  * Array  - actions (action (ADD|REPLACE|DELETE), key, value, ...)

Returns:
  * Scalar - result code

=cut

sub modify
{
	my $self = shift;
	my $dn = shift;
	my @list = @_;
	utf8::decode($_) foreach @list;
	$self->_debug( q/Modify: dn="%s", |%s|/, $dn, join( '|', map { defined $_ ? $_ : '<undef>' } @list ) ) if $self->verbose;

	my @actions = $self->_parse_actions( @list );

	my $entry = $self->store->entry( $dn );
	return LDAP_NO_SUCH_OBJECT unless defined $entry;

	foreach my $action ( @actions )
	{
		if( $action->{attr} ~~ [ $self->key_attribute ] )
		{
			return LDAP_NOT_ALLOWED_ON_RDN unless $action->{action} eq 'REPLACE'
				&& @{$action->{values}} == 1 && $action->{values}->[0] eq $entry->get_value($action->{attr});
		}
	}

	if( $self->_is_entry_stored( $entry ) )
	{
		my( $code, @entries ) = $self->_search( entries => [ $entry ] );
		return $code if $code != LDAP_SUCCESS;
		$entry = shift @entries;
		return LDAP_NO_SUCH_OBJECT unless defined $entry;
	}

	my $new_entry = $entry->clone();
	foreach my $action ( @actions )
	{
		if( $action->{action} eq 'ADD' )
		{
			$new_entry->add( $action->{attr} => $action->{values} );
		}
		elsif( $action->{action} eq 'REPLACE' )
		{
			$new_entry->replace( $action->{attr} => $action->{values} );
		}
		elsif( $action->{action} eq 'DELETE' )
		{
			$new_entry->delete( $action->{attr} => $action->{values} );
		}
	}

	if( $self->_is_entry_stored( $new_entry ) )
	{
		if( $self->_is_entry_stored( $entry ) )
		{
			my $code = $self->_modify(
				entry   => $entry,
				actions => [ grep { $self->is_attr_stored( $_->{attr} ) } @actions ]
			);
			return $code if $code;
		}
		else
		{
			my $code = $self->_add( entry => $new_entry );
			return $code if $code;
		}
	}
	else
	{
		if( $self->_is_entry_stored( $entry ) )
		{
			my $code = $self->_delete( entry => $entry );
			return $code if $code;
		}
	}

	$self->store->save( $new_entry );

	return LDAP_SUCCESS;
}

=item B<add>( $entry_str )

Purpose:
  * Add

Arguments:
  * Scalar - entry string

Returns:
  * Scalar - result code

=cut

sub add
{
	my $self = shift;
	my $entry_str = shift;
	$self->_debug( q/Add: "%s"/, $entry_str ) if $self->verbose;

	my $fh = IO::String->new( \$entry_str );
	my $entry = Net::LDAP::LDIF->new( $fh, 'r', raw => qr/;binary/ )->read_entry();
	$fh->close();

	if( $self->_is_entry_stored( $entry ) )
	{
		my $code = $self->_add( entry => $entry );
		return $code if $code;
	}

	$self->store->save( $entry );

	return LDAP_SUCCESS;
}

=item B<modrdn>( $dn, $new_dn, $del_flag )

Purpose:
  * Modify DN

Arguments:
  * Scalar - DN
  * Scalar - new RDN
  * Scalar - delete old DN flag

Returns:
  * Scalar - result code

=cut

sub modrdn
{
	my $self = shift;
	my( $dn, $new_rdn, $del_old, $new_sup ) = @_;
	$self->_debug( q/Modrdn: dn="%s", new_rdn="%s", new_sup="%s", del_old=%s/,
		$dn, $new_rdn, $new_sup // '<undef>', $del_old ? 1 : 0 ) if $self->verbose;

	my $entry = $self->store->entry( $dn );
	return LDAP_NO_SUCH_OBJECT unless defined $entry;

	my( $old_rdn ) = $entry->dn() =~ /^(.*?),/;
	# We can store only one key in DB
	return LDAP_UNWILLING_TO_PERFORM if !$del_old && $new_rdn ne $old_rdn;

	my $new_dn = sprintf "$new_rdn,%s", defined $new_sup ? $new_sup : $entry->dn() =~ /,(.*)$/;

	my( $attr, $value ) = split /\s*=\s*/, $new_rdn;
	$value = $self->_unescape( $value );

	my( @add, @del, @save );
	foreach my $subentry ( $self->store->entries( $entry->dn(), LDAP_SCOPE_SUB ) )
	{
		my $newsubentry = $subentry->clone();
		if( $self->_is_entry_stored( $subentry ) )
		{
			push @del, $subentry;
			my( $code, @entries ) = $self->_search( entries => [ $newsubentry ] );
			return $code if $code;
			$newsubentry = shift @entries;
		}
		my $sdn = $newsubentry->dn();
		$newsubentry->replace( $attr => $value ) if $sdn eq $dn;
		substr $sdn, index( $sdn, $dn ), length( $dn ), $new_dn;
		$newsubentry->dn( $sdn );
		push @add, $newsubentry if $self->_is_entry_stored( $newsubentry );
		$subentry->delete();
		push @save, $subentry, $newsubentry;
	}

	foreach my $subentry ( sort { length $b->dn() <=> length $a->dn() } @del )
	{
		my $code = $self->_delete( entry => $subentry );
		return $code if $code;
	}
	foreach my $subentry ( sort { length $a->dn() <=> length $b->dn() } @add )
	{
		my $code = $self->_add( entry => $subentry );
		return $code if $code;
	}

	$self->store->save( @save );

	return LDAP_SUCCESS;
}

=item B<delete>( $dn )

Purpose:
  * Delete

Arguments:
  * Scalar - DN

Returns:
  * None

=cut

sub delete
{
	my $self = shift;
	my $dn = shift;
	$self->_debug( q/Delete: dn="%s"/, $dn ) if $self->verbose;

	my $entry = $self->store->entry( $dn );
	return LDAP_NO_SUCH_OBJECT unless defined $entry;

	if( $self->_is_entry_stored( $entry ) )
	{
		my( $code, @entries ) = $self->_search( entries => [ $entry ] );
		return $code if $code != LDAP_SUCCESS;
		$entry = shift @entries;
		return LDAP_NO_SUCH_OBJECT unless defined $entry;

		$code = $self->_delete( entry => $entry );
		return $code if $code != LDAP_SUCCESS;
	}

	$entry->delete();
	$self->store->save( $entry );

	return LDAP_SUCCESS;
}

=item B<config>( @args )

Purpose:
  * Config

Arguments:
  * Array - arguments

Returns:
  * Scalar - result code

=cut

sub config
{
	my $self = shift;
	my( $name, @values ) = @_;
	$self->_debug( q/Config: name="%s"/ . q/, value="%s"/ x @values, $name, @values );
	if( $name eq 'proxyStore' )
	{
		$self->{store_mod} = shift @values;
		$self->{store_opt} = \@values;
		return LDAP_SUCCESS;
	}
	else
	{
		return $self->_config( $name => @values );
	}
}

=back

=head1 PROTECTED METHODS

=over

=item B<is_attr_stored>( $attr )

Purpose:
  * Checks if this C<$attr> is stored in DB

Arguments:
  * Scalar - attribute name

Returns:
  * Bool - is stored

=cut

sub is_attr_stored
{
	my( $self, $attr ) = @_;
	return any { $attr eq $_ } $self->stored_attributes;
}

=item B<parse_dn>( $dn )

Purpose:
  * Parse DN

Arguments:
  * Scalar - DN

Returns:
  * Array - key-value pairs of DN

=cut

sub parse_dn
{
	my( $self, $dn ) = @_;
	my @parts = split /\s*,\s*/, $dn;
	my @dn;
	foreach my $part ( @parts )
	{
		my( $attr, $value ) = split /=/, $part, 2;
		$value = $self->_unescape( $value );
		push @dn, $attr, $value;
	}
	return @dn;
}

=item B<parse_filter>( $filter_str )

Purpose:
  * Parse filter

Arguments:
  * Scalar - filter string

Returns:
  * Hashref - parsed filter

=cut

sub parse_filter
{
	my( $self, $filter_str ) = @_;
	my %filter;
	if( $filter_str =~ /^\(([^=]+)=([^\)]+)\)$/ )
	{
		$filter{$1} = $self->_unescape( $2 );
	}
	elsif( $filter_str =~ /^\(&((?:\([^=]+=[^\)]+\))+)\)$/ )
	{
		my $and_filter = $1;
		while( $and_filter =~ /\(([^=]+)=([^\)]+)\)/g )
		{
			$filter{$1} = $self->_unescape( $2 );
		}
	}
	return \%filter;
}

=item B<check_filter>( $entry, $filter )

Purpose:
  * Check if entry is match to filter

Arguments:
  * Net::LDAP::Entry - entry
  * Hashref - filter

Returns:
  * Scalar - boolean

=cut

sub check_filter
{
	my( $self, $entry, $filter ) = @_;
	my $match = 1;
	foreach my $attr ( keys %$filter )
	{
		$match = undef unless $entry->exists( $attr )
			&& ( $filter->{$attr} eq '*' || any { lc $_ eq lc $filter->{$attr} } $entry->get_value( $attr ) );
	}
	return $match;
}

=back

=head1 PRIVATE METHODS

=over

=item B<_is_entry_stored>( $entry )

Purpose:
  * Checks if this C<$entry> is stored in DB

Arguments:
  * Net::LDAP::Entry - entry

Returns:
  * Bool - is stored

=cut

sub _is_entry_stored
{
	my( $self, $entry ) = @_;

	foreach my $type ( $self->stored_entries )
	{
		my $method = "_is_$type";
		return 1 if $self->$method( $entry );
	}

	return;
}

=item B<_unescape>( $str )

Purpose:
  * Unescape attribute value

Arguments:
  * Scalar - escaped attribute value

Returns:
  * Scalar - attribute value

=cut

sub _unescape
{
	my( $self, $str ) = @_;
	$str =~ s/\\([0-9A-Fa-f]{2})/chr( hex( $1 ) )/eg;
	utf8::decode($str);
	return $str;
}

=item B<_parse_actions>( @list )

Purpose:
  * Parse actions

Arguments:
  * Scalar - actions (action (ADD|REPLACE|DELETE), attr, value, ...)

Returns:
  * Array(ref) - actions

=cut

sub _parse_actions
{
	my $self = shift();
	my @list = @_;

	my @actions;
	while( my $value = shift @list )
	{
		if( any { $value eq $_ } qw/ADD REPLACE DELETE/ )
		{
			push @actions, {
				action => $value,
				attr   => shift( @list ),
				values => [],
			};
		}
		else
		{
			push @{$actions[-1]->{values}}, $value;
		}
	}

	return wantarray ? @actions : \@actions;
}

=item B<_log>( $level, $format, @args )

Purpose:
  * Write message to syslog

Arguments:
  * Scalar - log level
  * Scalar - log message format
  * Array  - message arguments

Returns:
  * None

=cut

sub _log
{
	my $self = shift;
	my( $level, @args ) = @_;
	my $error = $!;

	utf8::encode($_) foreach @args;

	my $mask = shift @args;
	if( $mask =~ /%m/ )
	{
		# escape percent signs for sprintf()
		$error =~ s/%/%%/g if @_;
		# replace %m with $error, if preceded by an even number of percent signs
		$mask =~ s/(?<!%)((?:%%)*)%m/$1$error/g;
	}

	my $message = @args ? sprintf( $mask, @args ) : $mask;

	# Syslog is already opened by slapd, so we directly use this function
	Sys::Syslog::syslog_xs( $level, $message );

	return;
}

=item B<_error>( $format, @args )

Purpose:
  * Write error message to syslog

Arguments:
  * Scalar - log message format
  * Array  - message arguments

Returns:
  * None

=cut

sub _error
{
	my $self = shift;
	$self->_log( LOG_ERR, 'error: '.shift, @_ );
	return;
}

=item B<_warning>( $format, @args )

Purpose:
  * Write warning message to syslog

Arguments:
  * Scalar - log message format
  * Array  - message arguments

Returns:
  * None

=cut

sub _warning
{
	my $self = shift;
	$self->_log( LOG_WARNING, 'warning: '.shift, @_ );
	return;
}

=item B<_info>( $format, @args )

Purpose:
  * Write informational message to syslog

Arguments:
  * Scalar - log message format
  * Array  - message arguments

Returns:
  * None

=cut

sub _info
{
	my $self = shift;
	$self->_log( LOG_INFO, 'info: '.shift, @_ );
	return;
}

=item B<_debug>( $format, @args )

Purpose:
  * Write debug message to syslog

Arguments:
  * Scalar - log message format
  * Array  - message arguments

Returns:
  * None

=cut

sub _debug
{
	my $self = shift;
	$self->_log( LOG_DEBUG, 'debug: '.shift, @_ );
	return;
}

sub _sig_warn
{
	my $message = shift;
	__PACKAGE__->_warning( $message );
	return;
}

sub _sig_die
{
	my $message = shift;
	__PACKAGE__->_error( $message );
	return;
}

=back

=cut

1;

=head1 SEE ALSO

L<slapd-perl>, L<Net::LDAP>

=head1 AUTHOR

Aleksey Mashanov <aleksey.mashanov@gmail.com>

=head1 LICENSE

This library is free software, you can redistribute it and/or modify it under
the same terms as Perl itself.

=cut
