=head1 NAME

Slapd::Backend::Proxy::Store::BerkeleyDB - store intermediate data in BerkeleyDB

=cut

package Slapd::Backend::Proxy::Store::BerkeleyDB;

use utf8;
use strict;
use warnings;
use base 'Slapd::Backend::Proxy::Store';
use BerkeleyDB;
use IO::String;
use Net::LDAP::LDIF;

use Slapd::Backend::Constant;

__PACKAGE__->mk_accessors(qw( env db uuid_db ));

=head1 PUBLIC METHODS

=over

=item B<save>( @entries )

Purpose:
  * Save entries to disk

Arguments:
  * List - entries

Returns:
  * None

=cut

sub save
{
	my $self = shift;

	foreach my $entry ( @_ )
	{
		if( $entry->changetype() eq 'delete' )
		{
			my $status = $self->db->db_del( $entry->dn() );
			warn "$status ($BerkeleyDB::Error)" unless $status ~~ [0, DB_NOTFOUND];
		}
		else
		{
			my $str = new IO::String;
			my $ldif = new Net::LDAP::LDIF( $str, 'w', raw => qr/;binary/ );
			# write_entry() converts string to UTF-8 octets in-place, so we ought to clone entry.
			$ldif->write_entry( $entry->clone() );
			my $ldif_str = ${$str->string_ref()};
			$ldif_str =~ s/^\n//;

			my $status = $self->db->db_put( $entry->dn(), $ldif_str );
			warn "$status ($BerkeleyDB::Error)" unless $status == 0;
		}
	}

	return;
}

=item B<entries>( $base, $scope, $uuid )

Purpose:
  * Get entries

Arguments:
  * Scalar - base
  * Scalar - scope
  * Scalar - entry uuid (optional)

Returns:
  * List of Net::LDAP::Entry

=cut

sub entries
{
	my( $self, $base, $scope, $uuid ) = @_;

	my @ldif;
	if( $scope == LDAP_SCOPE_BASE )
	{
		my $ldif;
		my $status = $self->db->db_get( $base, $ldif );
		warn "$status ($BerkeleyDB::Error)" unless $status ~~ [0, DB_NOTFOUND];
		push @ldif, $ldif if $status == 0;
	}
	elsif( $uuid )
	{
		my $ldif;
		my $status = $self->uuid_db->db_get( $uuid, $ldif );
		warn "$status ($BerkeleyDB::Error)" unless $status ~~ [0, DB_NOTFOUND];
		push @ldif, $ldif if $status == 0;
	}
	else
	{
		my $regexp = $scope == LDAP_SCOPE_ONE ? qr/^[^,]+,$base$/
			: $scope == LDAP_SCOPE_SUB ? qr/^(?:.+,)?$base$/
			: $scope == LDAP_SCOPE_CHILDREN ? qr/^.+,$base$/
			: undef;
		my $cursor = $self->db->db_cursor() or die $BerkeleyDB::Error;
		my $status;
		my $dn = '';
		my $ldif = '';
		while( ( $status = $cursor->c_get( $dn, $ldif, DB_NEXT ) ) == 0 )
		{
			next unless $dn =~ $regexp;
			push @ldif, $ldif;
		};
		warn "$status ($BerkeleyDB::Error)" unless $status ~~ [0, DB_NOTFOUND];
		$cursor->c_close();
	}
	my @entries;
	foreach my $ldif ( @ldif )
	{
		my $fh = IO::String->new( \$ldif );
		my $entry = Net::LDAP::LDIF->new( $fh, 'r', raw => qr/;binary/ )->read_entry();
		$fh->close();
		next if $uuid && $entry->get_value('entryUUID') ne $uuid;
		push @entries, $entry;
	}

	return @entries;
}

=back

=head1 PROTECTED METHODS

=over

=item B<_init>( $filename )

Purpose:
  * Initialization

Arguments:
  * Scalar - filename

Returns:
  * None

=cut

sub _init
{
	my $self = shift;
	my( $home ) = @_;

	my $env = new BerkeleyDB::Env(
		-Home  => $home,
		-Flags => DB_CREATE| DB_INIT_CDB | DB_INIT_MPOOL,
	);

	my $db = new BerkeleyDB::Hash(
		-Env      => $env,
		-Filename => "ldif.db",
		-Flags    => DB_CREATE,
	) or die "Can't open file ldif.db: $! $BerkeleyDB::Error\n";

	my $sdb = new BerkeleyDB::Hash(
		-Env      => $env,
		-Filename => "uuid.db",
		-Flags    => DB_CREATE,
	) or die "Can't open file uuid.db: $! $BerkeleyDB::Error\n";

	$db->associate(
		$sdb,
		sub {
			my( $key, $data ) = @_;
			my $fh = IO::String->new( \$data );
			my $entry = Net::LDAP::LDIF->new( $fh, 'r', raw => qr/;binary/ )->read_entry();
			$fh->close();
			if( my $uuid = $entry->get_value('entryUUID') )
			{
				$_[2] = $uuid; # Third arg is $skey
				return 0;
			}
			else
			{
				return DB_DONOTINDEX;
			}
		}
	);

	$self->env = $env;
	$self->db = $db;
	$self->uuid_db = $sdb;

	return;
}

=back

=cut

1;

=head1 SEE ALSO

L<Slapd::Backend::Proxy::Store>, L<Slapd::Backend::Proxy>, L<BerkeleyDB>

=head1 AUTHOR

Aleksey Mashanov <aleksey.mashanov@gmail.com>

=head1 LICENSE

This library is free software, you can redistribute it and/or modify it under
the same terms as Perl itself.

=cut
